'''Contains ProblemDefinition'''
import json
import textwrap
from typing import Any, Dict, List, Optional, Union

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd
from pandas import DataFrame

from ..dataset_formats.impl.dataset_config import (DatasetConfig,
                                                   DatasetFileError,
                                                   DatasetConfigError)
from ..dataset_formats.impl.dataset_format_catalog import (DatasetFormatCatalog,
                                                           DatasetFormatCatalogAuto)
from ..executor.executor_kind import ExecutorKind
from ..wrangler.constants import ProblemDefKeys, ProblemDefKeySet, Defaults
from ..wrangler.dataset import Metadata, RoleName

from .aggregation_config import AggregationConfig, AggregationError
from .cross_validation_config import CrossValidationConfig
from .config_component import ConfigComponent, ConfigError, ParsingErrors
from .config_component import MissingKeyError
from .config_component import InvalidKeyError, ValidationErrors
from .config_component import InvalidValueError, ProblemDefTypeError
from .hyperparam_config import HyperparamConfig, HyperparamError
from .metric_config import MetricConfig, MetricConfigError
from .output_config import OutputConfig
from .task import TaskType, Task

# pylint: disable=too-many-branches,too-many-statements


class ProblemDefinition(ConfigComponent):
    '''Represents information needed to define an AutoML run'''
    _task: Optional[Task] = None

    _aggregation_config: Optional[AggregationConfig] = None
    _cross_validation_config: Optional[CrossValidationConfig] = None
    _dataset_config: Optional[DatasetConfig] = None
    _metric_configs: Optional[Dict[str, MetricConfig]] = None
    _hyperparam_config: Optional[HyperparamConfig] = None
    _output_config: Optional[OutputConfig] = None

    _data_formats_catalog: Optional[DatasetFormatCatalog] = None

    def __init__(self,
                 problem_def: Union[str, Dict[str, Any]],
                 train_df: Optional[DataFrame] = None,
                 test_df: Optional[DataFrame] = None,
                 data_formats_catalog: Optional[DatasetFormatCatalog] = None):

        if isinstance(problem_def, str):
            problem_def = self._parse_json(pdef=problem_def)
        super().__init__(clause=remove_comments(problem_def))
        self._data_formats_catalog = data_formats_catalog or DatasetFormatCatalogAuto()

        errors: List[ConfigError] = []

        try:
            self._task = Task(self._get(
                ProblemDefKeys.TASK))
        except KeyError as err:
            errors.append(InvalidKeyError(err))

        try:
            clause = self._get(ProblemDefKeys.DATASET)
            self._dataset_config = self._data_formats_catalog.build(
                clause=clause,
                task=self._task,
                train_df=train_df,
                test_df=test_df
            )
        except (KeyError,  # Missing field names, bad column name, bad role name
                FileNotFoundError,  # Missing csv
                TypeError,  # Malformed json.
                pd.errors.ParserError,  # Train file is not parsable.
                IndexError  # Column id out of range.
                ) as err:
            errors.append(DatasetConfigError(err))
        except DatasetConfigError as err:  # Train dataframe not provided for MemoryDatasetConfig
            errors.append(err)

        try:
            self._cross_validation_config = CrossValidationConfig(self._get_with_default(
                ProblemDefKeys.CROSS_VALIDATION, dflt={}))
        except KeyError as err:
            errors.append(InvalidKeyError(err))

        try:
            metrics_clause = self._get_with_default(ProblemDefKeys.METRICS, dflt={})
            self._metric_configs = {name: MetricConfig(clause={name: info})
                                    for name, info in metrics_clause.items()}
        except MetricConfigError as err:
            errors.append(err)

        try:
            hyperparam_clause = self._get_with_default(ProblemDefKeys.HYPERPARAMS, dflt=[])
            self._hyperparam_config = HyperparamConfig(
                clause={ProblemDefKeys.HYPERPARAMS.value: hyperparam_clause})
        except HyperparamError as err:
            errors.extend(err.errors)

        try:
            output_clause = self._get_with_default(ProblemDefKeys.OUTPUT, dflt={})
            self._output_config = OutputConfig(output_clause)
            # TODO(piggy): Confirm that OutputConfig calls its own validate()
            # and drop this call.
            self._output_config.validate()
        except ValidationErrors as err:
            errors.extend(err.errors)

        try:
            aggregation_clause = self._get_with_default(ProblemDefKeys.AGGREGATION, dflt={})
            self._aggregation_config = AggregationConfig(aggregation_clause)
        except AggregationError as err:
            errors.extend(err.errors)

        try:
            # This is where we do validation between sections.
            self.validate()
        except ValidationErrors as err:
            errors.extend(err.errors)

        if len(errors) > 0:
            raise ValidationErrors(errors)

    def _parse_json(self, pdef: str) -> Dict[str, Any]:
        '''Attempts to interpret pdef as a JSON string or path to JSON file.

        If neither succeeds, raises ParsingErrors.
        '''
        errors: List[ConfigError] = []

        try:
            retval = json.loads(pdef)
            return retval
        except json.decoder.JSONDecodeError as err:
            errors.append(ConfigError(err))

        try:
            with open(pdef, 'r', encoding='utf8') as f:
                retval = json.load(f)
            return retval
        except FileNotFoundError as err:
            errors.append(ConfigError(err))
        except json.decoder.JSONDecodeError as err:
            errors.append(ConfigError(err))

        errors.insert(0, ConfigError(
            'String provided could not be parsed as JSON or path to JSON file.'))
        raise ParsingErrors(errors=errors)

    def __str__(self):
        return textwrap.dedent(f'''
            problem_type={self.task}
            dataset_config={self.dataset_config}
        ''')

    def validate(self, **kwargs) -> None:
        errors: List[ConfigError] = []
        keys = set(self._clause.keys())
        if not ProblemDefKeySet.REQUIRED.issubset(keys):
            errors.append(MissingKeyError(
                'Required keys missing in problem definition: '
                f'{ProblemDefKeySet.REQUIRED.difference(keys)}'))
        if not keys.issubset(ProblemDefKeySet.ALLOWED):
            errors.append(InvalidKeyError(
                'Invalid key(s) in problem definition: '
                f'{keys.difference(ProblemDefKeySet.ALLOWED)}'))

        if self._task is not None:
            try:
                self.task.validate()
            except ValidationErrors as err:
                errors.extend(err.errors)

        if self._dataset_config is not None:
            try:
                self.dataset_config.validate()
            except ValidationErrors as err:
                errors.extend(err.errors)

        if self._task is not None and self._dataset_config is not None:
            try:
                self._validate_task_dataset()
            except ValidationErrors as err:
                errors.extend(err.errors)
            except ConfigError as err:
                errors.append(err)

        if self._metric_configs is not None:
            for metric_config in self._metric_configs.values():
                try:
                    metric_config.validate()
                except MetricConfigError as err:
                    errors.append(err)

        if self._cross_validation_config is not None \
                and self._dataset_config is not None \
                and self._task is not None:
            # we need information about the dataset in order to sanity check
            #   cross-validation settings (i.e., 0 < k <= total number of rows in dataframe)
            try:
                pass
                # train_df = self._dataset_config.load_train().dataframe
                # self._cross_validation_config.validate(df_row_count=train_df.shape[0])
            except ProblemDefTypeError as err:
                errors.append(err)
            except InvalidValueError as err:
                errors.append(err)

        if len(errors) > 0:
            raise ValidationErrors(errors)

    def _validate_task_dataset(self) -> None:
        errors: List[ConfigError] = []
        if self.task.task_type == TaskType.FORECASTING:
            if RoleName.TIME not in self.dataset_config.roles:
                errors.append(
                    DatasetFileError(
                        'Dataset for forecasting problem must have a time column.\n'
                        f'Found roles: {self.dataset_config.roles!r}'
                    )
                )

            if self.dataset_config.forecasting is None:
                errors.append(
                    MissingKeyError(
                        'Dataset config for forecasting problem must have a forecasting clause.'
                    )
                )

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)

    @property
    def dataset_config(self) -> DatasetConfig:
        '''The dataset description.'''
        assert self._dataset_config is not None, (
            'BUG: attempt to access dataset_config but it is None.')
        return self._dataset_config

    @property
    def task(self) -> Task:
        '''Explains what kind of problem we're solving.

        This affects the selection of pipeline templates and models.
        '''
        assert self._task is not None, (
            'BUG: attempt to access problem_type but it is None.')
        return self._task

    @property
    def metric_configs(self) -> Dict[str, MetricConfig]:
        '''Names and other information, if applicable, about all the metrics for this problem.

        If there are no metrics specified, will return an empty list, in which case the wrangler is
        responsible for using a default metric based on the problem type.
        '''
        assert self._metric_configs is not None, (
            'BUG: attempt to access metric_config but it is None.')
        return self._metric_configs

    @property
    def cross_validation_config(self) -> CrossValidationConfig:
        '''
        Get configuration information for cross-validation.
        '''
        assert self._cross_validation_config is not None, (
            'BUG: attempt to access cross_validation_config but it is None.'
        )

        return self._cross_validation_config

    @property
    def instantiations(self) -> List[ExecutorKind]:
        '''The user requests these kinds of output pipeline instantiations.

        A missing instantions clause defaults to all available kinds.
        An empty instantions clause disables instantiation.
        '''
        inst = self._get_with_default(ProblemDefKeys.OUTPUT, ProblemDefKeys.INSTATIATIONS,
                                      dflt=Defaults.INSTANTIANTIONS)

        return [ExecutorKind(i.lower()) for i in inst]

    @property
    def output(self) -> OutputConfig:
        '''All the configuration about things to output.'''
        assert self._output_config is not None, (
            'BUG: empty output session should be handled in constructor.'
        )
        return self._output_config

    @property
    def metadata(self) -> Metadata:
        '''Package config metadata needed by some models.'''
        return self.dataset_config.metadata

    @property
    def hyperparams(self) -> HyperparamConfig:
        '''Configuration for all hyperparam overrides.'''
        assert self._hyperparam_config is not None, (
            'BUG: self._hyperparam_config is supposed to be set in __init__().')
        return self._hyperparam_config

    @property
    def aggregation(self) -> AggregationConfig:
        '''Configuration for rank aggregation.

        The only valid entry is "method" which gives
        the catalog name of a rank aggregation method.
        See the top level "aggregators" directory.
        '''
        assert self._aggregation_config is not None, (
            'BUG: self._aggregation_config is supposed to be set in __init__().')
        return self._aggregation_config


def remove_comments(entry: Any) -> Any:
    '''Remove '_comments' fields from every dict inside an anonymous object.'''
    if isinstance(entry, dict):
        retval = {
            k: remove_comments(v)
            for k, v in entry.items()
            if k != '_comments'
        }
        return retval
    if isinstance(entry, list):
        return [remove_comments(v) for v in entry]
    return entry



'''Test for dataframe_dataset_config.py'''

# pylint: disable=missing-function-docstring,missing-class-docstring

from typing import Optional

import pandas as pd

from ...wrangler.dataset import Dataset, RoleName
from .dataframe_dataset_config import DataframeDatasetConfig


class FakeConfig(DataframeDatasetConfig):

    def load_train(self) -> Dataset:
        return Dataset()

    def load_test(self) -> Optional[Dataset]:
        return None


def test_dataset_cols() -> None:
    dut = FakeConfig({})
    got = dut.dataset(
        pd.DataFrame({
            'a': [1, 2, 3],
            'b': [4, 5, 6]
        }),
        cols=['a']
    )

    want = pd.DataFrame({
        'a': [1, 2, 3]
    })

    pd.testing.assert_frame_equal(want, got.dataframe)


def test_dataset_roles() -> None:
    dut = FakeConfig({
        'column_roles': {
            'target': {'name': 'a'},
            'test_role': {'name': 'b'},
            'attribute': {'name': 'c'},
        }
    })
    df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9]
    })
    got1 = dut.dataset(data=df, roles=[RoleName.TARGET])
    want1 = pd.DataFrame({
        'a': [1, 2, 3]})
    pd.testing.assert_frame_equal(got1.dataframe, want1)

    got2 = dut.dataset(data=df, roles=['attribute', RoleName.TEST_ROLE])
    want2 = pd.DataFrame({
        'b': [4, 5, 6],
        'c': [7, 8, 9],
    })
    pd.testing.assert_frame_equal(got2.dataframe, want2)


def test_dataset_roles_cols() -> None:
    dut = FakeConfig({
        'column_roles': {
            'target': {'name': 'a'},
            'attribute': {'name': 'c'},
            'test_role': {'name': 'd'},
        }
    })
    df = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9],
        'd': [10, 12, 13]  # ROFL
    })
    got = dut.dataset(data=df, roles=['target', 'attribute'], cols=['a', 'b'])
    want = pd.DataFrame({
        'a': [1, 2, 3],
        'b': [4, 5, 6],
        'c': [7, 8, 9]
    })
    pd.testing.assert_frame_equal(got.dataframe, want)



"""The wrangler is the central control object for all of AutonML."""
import logging
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Type, Union

from ..aggregators.impl.aggregate_ranker import AggregateRanker
from ..aggregators.impl.aggregator_auto import AggregatorCatalogAuto
from ..aggregators.impl.aggregator_catalog import AggregatorCatalog
from ..algorithms.impl.algorithm import AlgorithmCatalog
from ..algorithms.impl.algorithm_auto import AlgorithmCatalogAuto
from ..catalog.catalog import Catalog
from ..cross_validators.impl.cross_validator import CrossValidatorCatalog
from ..cross_validators.impl.cross_validator_auto import \
    CrossValidatorCatalogAuto
from ..executor.executor import Executor
from ..executor.simple.simple_executor import SimpleExecutor
from ..generator.bound_pipeline import BoundPipeline
from ..generator.designator import Designator
from ..generator.generator import Generator, GeneratorImpl
from ..instantiator.executable_pipeline import (ExecutablePipeline,
                                                PipelineResult,
                                                PipelineResults)
from ..instantiator.instantiator_factory import InstantiatorFactory
from ..metrics.impl.metric import Metric
from ..metrics.impl.metric_catalog import MetricCatalog
from ..metrics.impl.metric_auto import MetricCatalogAuto
from ..problem_def.problem_def import ProblemDefinition
from ..ranker.ranker import Ranker, Rankings
from ..ranker.ranker_impl import RankerImpl
from ..searcher.searcher import Searcher, SearcherImpl
from ..splitters.impl.splitter import SplitDataset, SplitterCatalog
from ..splitters.impl.splitter_auto import SplitterCatalogAuto
from ..templates.impl.pipeline_template import PipelineTemplate
from ..templates.impl.template import TemplateCatalog
from ..templates.impl.template_auto import TemplateCatalogAuto
from .dataset import Dataset
from .saver import Saver

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=too-many-instance-attributes,too-many-arguments


class Error(Exception):
    '''Base error for Wrangler.'''


class WranglerFailure(Error):
    '''Wrangler can not procede.'''


class ReassignmentError(Error):
    '''An attempt was made to assign a value more than once.'''


class WranglerResult():
    '''The results of Wrangler.wrangle.'''
    _split_dataset: SplitDataset
    _train_results: PipelineResults
    _test_results: Optional[PipelineResults]
    _rankings: Rankings

    def __init__(self,
                 split_dataset: SplitDataset,
                 train_results: PipelineResults,
                 test_results: Optional[PipelineResults],
                 rankings: Rankings):
        self._split_dataset = split_dataset
        self._train_results = train_results
        self._test_results = test_results
        self._rankings = rankings

    @property
    def split_dataset(self) -> SplitDataset:
        '''The ground truth and folds used to rank the pipelines.'''
        return self._split_dataset

    @property
    def train_results(self) -> PipelineResults:
        '''Predictions on train data, acquired using cross-validation.'''
        return self._train_results

    @property
    def test_results(self) -> Optional[PipelineResults]:
        '''Predictions on test data, if supplied in the problem definition.'''
        return self._test_results

    @property
    def rankings(self) -> Rankings:
        '''The rankings of all bound pipelines.'''
        return self._rankings

    @property
    def executable_pipelines(self) -> Dict[Designator, ExecutablePipeline]:
        '''Get all the executable pipelines.'''
        return self._train_results.executable_pipelines


class Wrangler():
    '''The wrangler is the central control object for all of AutonML.'''
    _bound_pipelines: Optional[Dict[Designator, BoundPipeline]] = None
    # These are all pipelines identified during the first run.
    _all_pipelines: Optional[Dict[Designator, ExecutablePipeline]] = None
    # This is a subset of _all_pipelines which will be the default
    # for subsequent fit, predict, and rank methods.
    _current_pipelines: Optional[Dict[Designator, ExecutablePipeline]] = None

    def __init__(
        self,
        problem_definition: ProblemDefinition,
        template_catalog: Optional[Type[TemplateCatalog]] = None,
        metric_catalog: Optional[Type[MetricCatalog]] = None,
        algorithm_catalog: Optional[Type[AlgorithmCatalog]] = None,
        generator: Optional[Type[Generator]] = None,
        searcher: Optional[Type[Searcher]] = None,
        executor: Optional[Type[Executor]] = None,
        ranker: Optional[Type[Ranker]] = None,
        instantiator_factory: Optional[Type[InstantiatorFactory]] = None,
        splitter_catalog: Optional[Type[SplitterCatalog]] = None,
        validator_catalog: Optional[Type[CrossValidatorCatalog]] = None,
        saver: Optional[Type[Saver]] = None,
        aggregator_catalog: Optional[Type[AggregatorCatalog]] = None,
    ):
        # TODO(Merritt): initialize the logger
        # instantiate default components only if input is None
        self._pd = problem_definition
        self._metric_catalog = (metric_catalog or MetricCatalogAuto)()
        self._algorithm_catalog = (algorithm_catalog or AlgorithmCatalogAuto)()
        self._generator = (generator or GeneratorImpl)(
            algorithm_catalog=self._algorithm_catalog,
            problem_definition=self._pd)
        self._template_catalog = (template_catalog or TemplateCatalogAuto)(
            algorithm_catalog=self._algorithm_catalog,
            generator=self._generator
        )
        self._searcher = (searcher or SearcherImpl)(hyperparams=self._pd.hyperparams)
        self._executor = (executor or SimpleExecutor)()
        self._ranker = (ranker or RankerImpl)()
        self._splitter_catalog = (splitter_catalog or SplitterCatalogAuto)(
            cv_config=self._pd.cross_validation_config)
        self._validator_catalog = (validator_catalog or CrossValidatorCatalogAuto)()
        self._aggregator_catalog = (aggregator_catalog or AggregatorCatalogAuto)()
        self._saver = None
        # If there is no output path, we'll output no files.
        if self._pd.output.path is not None:
            self._saver = (saver or Saver)(self._pd.output)
        self._instantiator_factory = (
            instantiator_factory or InstantiatorFactory)(
                saver=self._saver)

    @property
    def aggregator_catalog(self) -> AggregatorCatalog:
        '''Query and register aggregators'''
        return self._aggregator_catalog

    @property
    def algorithm_catalog(self) -> AlgorithmCatalog:
        '''Query and register algorithms'''
        return self._algorithm_catalog

    @property
    def metric_catalog(self) -> Catalog[Metric]:
        '''Query and register metrics'''
        return self._metric_catalog

    @property
    def template_catalog(self) -> TemplateCatalog:
        '''Query and register templates'''
        return self._template_catalog

    @property
    def generator(self) -> Generator:
        '''Generates bound pipelines from templates.'''
        return self._generator

    @property
    def ranker(self) -> Ranker:
        '''Ranks the results of running pipelines.'''
        return self._ranker

    def load_train_dataset(self) -> Dataset:
        '''Load self._pd.dataset()'''
        return self._pd.dataset_config.load_train()

    def load_test_dataset(self) -> Optional[Dataset]:
        '''Load self._pd.dataset()'''
        return self._pd.dataset_config.load_test()

    def dataset(self, data: Any, **kwargs) -> Dataset:
        '''Load in-memory data into a Dataset object, stored at key 'key'.'''
        return self._pd.dataset_config.dataset(data, **kwargs)

    def lookup_templates(self) -> Dict[str, PipelineTemplate]:
        '''Look up the templates that match the problem definition.'''
        task = self._pd.task
        data_type = "None"
        if task.data_type is not None:
            data_type = task.data_type.name
        task_type = "None"
        if task.task_type is not None:
            task_type = task.task_type.name

        return self._template_catalog.lookup_by_both(
            data_type=[data_type], task=[task_type])

    def train_all_results(self,
                          results: PipelineResults,
                          dataset: Dataset) -> PipelineResults:
        '''Instantiate and train all pipelines in results.

        Note that we do NOT calculate new predictions.
        '''
        retval = PipelineResults()
        pipelines: Dict[Designator, ExecutablePipeline] = {}

        for des, result in results.items():
            pipelines[des] = self._instantiator_factory.instantiate(
                kind=self._executor.kind,
                pipeline=result.bound_pipeline)

        self._executor.fit(
            dataset=dataset,
            pipelines=pipelines)

        for des, pipe in pipelines.items():
            retval[des] = PipelineResult(
                prediction=results[des].prediction,
                split_dataset=results[des].split_dataset,
                executable_pipeline=pipe)

        return retval

    def _predict_test_data(self,
                           executable_pipelines: Dict[Designator, ExecutablePipeline]
                           ) -> Optional[PipelineResults]:
        '''Run the executor again on test data to get test predictions,
            if test data is supplied in the problem definition.

        (Currently runs all pipelines, may eventually run x best ones)'''
        assert self._bound_pipelines is not None, (
            'BUG: wrangle_test() called with no bound pipelines.')

        test_dataset = self.load_test_dataset()

        if test_dataset is None:
            return None

        test_predictions = self._executor.predict(
            dataset=test_dataset,
            pipelines=executable_pipelines)
        return PipelineResults(test_predictions)

    def fit_predict_rank(self) -> WranglerResult:
        '''Do all the autoML things.'''
        train_dataset = self.load_train_dataset()

        templates = self.lookup_templates()
        if len(templates) == 0:
            raise WranglerFailure(f'found no templates for {self._pd.task}')
        logging.info('Found %d templates', len(templates))
        print(f'Found {len(templates)} templates')

        gen_result = self._generator.generate_all(templates)
        self._bound_pipelines = self._searcher.bind_all(gen_result)
        logging.info('Generated %d bound pipelines', len(self._bound_pipelines))
        print(f'Generated {len(self._bound_pipelines)} bound pipelines')

        task = self._pd.task
        assert task.task_type is not None, (
            'BUG: missing task should have been caught in validation.')
        splitters = self._splitter_catalog.lookup_by_tag_and(**{
            'task': task.task_type.name,
            'data_type': task.data_type.name,
        })
        if not splitters:
            splitters = self._splitter_catalog.lookup_by_tag_and(**{
                'default': 'true'
            })
        assert len(splitters) == 1, f'BUG: More than one splitter returned for {task}: {splitters}'
        splitter = list(splitters.values())[0]
        split_data = splitter.split(
            dataset=train_dataset,
            train_frac=self._pd.dataset_config.train_fraction,
            **self._pd.cross_validation_config.splitter_hyperparams
        )
        logging.info('Split dataset into %d folds.', len(split_data.folds))
        print(f'Split dataset into {len(split_data.folds)} folds.')

        cross_validator = self._validator_catalog.lookup_by_name('k_fold_cross_validator')
        cross_validator_results = PipelineResults(cross_validator.validate_pipelines(
            split_dataset=split_data,
            bound_pipelines=self._bound_pipelines,
            instantiator=self._instantiator_factory,
            executor=self._executor))

        metrics = self._metric_catalog.lookup_metrics(self._pd)
        print(f'Got {len(metrics)} metrics.')
        rankings = self._ranker.rank(
            results=cross_validator_results,
            metrics=metrics,
            ground_truth=split_data.ground_truth)
        logging.info('Rankings: %s', [str(rank) for rank in rankings])

        methods = [self._aggregator_catalog.lookup_by_name(method)
                   for method in self._pd.aggregation.method]
        new_rankings = AggregateRanker(methods=methods,
                                       rankings=rankings,
                                       results=cross_validator_results)()
        rankings.update(new_rankings)
        print(f'Added {len(new_rankings)} aggregate rankings.')

        train_results = self.train_all_results(
            results=cross_validator_results,
            dataset=train_dataset)

        if self._all_pipelines is None:
            self._all_pipelines = train_results.executable_pipelines
            self._current_pipelines = self._all_pipelines

        if self._saver is not None:
            self.save(train_results)

        test_results = self._predict_test_data(train_results.executable_pipelines)

        return WranglerResult(
            split_dataset=split_data,
            train_results=train_results,
            test_results=test_results,
            rankings=rankings)

    def save(self, train_results: PipelineResults) -> Dict[Designator, Path]:
        '''Save all the models and all the pipelines.'''
        assert self._saver is not None, 'BUG: Only call save if there is a saver.'
        retval: Dict[Designator, Path] = {}
        model_paths = self._saver.save_models(train_results.executable_pipelines)
        for kind in self._pd.instantiations:
            instantiatior = self._instantiator_factory.build(kind)
            for des, result in train_results.items():
                retval[des] = instantiatior.save(result.bound_pipeline, model_paths)
        return retval

    def fit(self,
            dataset: Dataset,
            pipelines: Optional[Dict[Designator, ExecutablePipeline]] = None
            ) -> Dict[Designator, ExecutablePipeline]:
        '''Fit the given pipelines with a new dataset.

        If no pipelines are specified, use the current set.
        '''
        if pipelines is None:
            pipelines = self._current_pipelines

        # Reinstantiate all the pipelines with clean models.
        assert pipelines is not None, (
            'BUG: fit called with None current pipelines'
        )
        for des, pipeline in pipelines.items():
            pipelines[des] = self._instantiator_factory.instantiate(
                kind=self._executor.kind,
                pipeline=pipeline.bound)

        self._executor.fit(
            dataset=dataset,
            pipelines=pipelines)

        return pipelines

    def predict(self,
                new_data: Dataset,
                trained_pipelines: Optional[
                    Dict[Designator, ExecutablePipeline]] = None) -> PipelineResults:
        '''Predict on new test data that was not supplied in the problem definition.'''
        if trained_pipelines is None:
            trained_pipelines = self._current_pipelines
        assert trained_pipelines is not None, (
            'BUG: predict called with None current pipelines.'
        )
        pipeline_errors: List[str] = []
        for des, pipe in trained_pipelines.items():
            if pipe.kind != self._executor.kind:
                pipeline_errors.append(
                    f'Pipeline {des} has incorrect ExecutorKind {pipe.kind}; '
                    f'expected {self._executor.kind}. \n')
            elif not pipe.trained:
                pipeline_errors.append(
                    f'Pipeline {des} is untrained. \n')

        if len(pipeline_errors) > 0:
            raise WranglerFailure(
                'Bad pipeline(s) provided to predict(): \n'
                f'{"".join(pipeline_errors)}')

        results = self._executor.predict(dataset=new_data, pipelines=trained_pipelines)
        return PipelineResults(results)

    def set_current(self, *args: str) -> None:
        '''Set the current set of pipelines to work with by name.

        If no arguments are given, reset to the full set of pipelines
        identified in the last fit_predict_rank.
        '''
        if len(args) == 0:
            self._current_pipelines = self._all_pipelines
            return
        if self._all_pipelines is None:
            self._current_pipelines = None
            return
        self._current_pipelines: Dict[Designator, ExecutablePipeline] = {
            Designator(des): self._all_pipelines[Designator(des)]
            for des in args
        }

    def rank(self,
             results: PipelineResults,
             metrics: Optional[Iterable[Union[str, Metric]]] = None,
             ground_truth: Optional[Dataset] = None) -> Rankings:
        '''Rank a set of pipeline results.

        If no metrics are provided, uses the ones from the problem definition

        If no ground truth is provided, attempts to infer it using the SplitDataset
        associated with an arbitrary PipelineResult.
        '''
        metrics_for_ranker: Dict[str, Metric] = {}
        if metrics is None:
            metrics_for_ranker = self._metric_catalog.lookup_metrics(self._pd)
        else:
            for m in metrics:
                if isinstance(m, Metric):
                    metrics_for_ranker[m.name] = m
                    continue
                metrics_for_ranker[m] = self._metric_catalog.lookup_by_name(m)

        if ground_truth is None:
            ground_truth = results.infer_ground_truth()

        return self.ranker.rank(results=results,
                                metrics=metrics_for_ranker,
                                ground_truth=ground_truth)



'''Metric configuration'''
from typing import Any, Dict, List, Optional

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..wrangler.constants import ProblemDefKeys
from .config_component import ConfigComponent, ConfigError


class MetricConfigError(ConfigError):
    '''Base error for MetricConfig.'''


class MetricConfig(ConfigComponent):
    '''Configured metrics'''
    _catalog_name: Optional[str] = None

    def __init__(self, clause: Dict[str, Any], parents: Optional[List[str]] = None):
        parents = self._add_parent(parents, ProblemDefKeys.METRICS.value)
        super().__init__(clause=clause, parents=parents)

    def validate(self, **kwargs) -> None:
        '''Throw a MetricConfigError if the metric clause is invalid'''
        errors: List[MetricConfigError] = []

        if len(self._clause.keys()) != 1:
            errors.append(MetricConfigError(
                'Metric clause must have exactly 1 key;'
                f'found {len(self._clause.keys())} ({self._clause.keys()})'
            ))

        try:
            # These are all run to make sure they generate no errors.
            self.name  # pylint: disable=pointless-statement
            self.hyperparams  # pylint: disable=pointless-statement
            self.catalog_name  # pylint: disable=pointless-statement
        except AssertionError as err:
            errors.append(MetricConfigError(str(err)))

        if len(errors) > 0:
            raise MetricConfigError(errors)

    @property
    def name(self) -> str:
        '''The instance name for this metric.'''
        return list(self._clause.keys())[0]

    @property
    def hyperparams(self) -> Dict[str, Any]:
        '''The hyperparams for this metric.'''
        retval = dict(list(self._clause.values())[0])
        if ProblemDefKeys.CATALOG_NAME.value in retval:
            del retval[ProblemDefKeys.CATALOG_NAME.value]
        return retval

    @property
    def catalog_name(self) -> str:
        '''The metric catalog name for this metric.

        Defaults to self.name.
        '''
        metric_info: Dict[str, Any] = list(self._clause.values())[0]
        retval = metric_info.get(ProblemDefKeys.CATALOG_NAME.value, self.name)
        assert isinstance(retval, str), f'catalog_name must be a str, not a {type(retval)}'
        return retval



'''Holds information about an in-memory dataset'''
from typing import Dict, Any, Optional, List

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import pandas as pd

from .impl.dataframe_dataset_config import DataframeDatasetConfig
from .impl.dataset_config import DatasetConfigError
from .impl.dataset_format_catalog import DatasetFormat, DatasetFormatCatalog
from ..wrangler.dataset import Dataset
from ..problem_def.task import Task


class MemoryDatasetConfig(DataframeDatasetConfig):
    ''' Holds information about an in-memory dataset '''

    # pylint: disable=too-many-arguments
    def __init__(self,
                 clause: Dict[str, Any],
                 train_df: Optional[pd.DataFrame] = None,
                 test_df: Optional[pd.DataFrame] = None,
                 parents: Optional[List[str]] = None,
                 task: Optional[Task] = None):
        super().__init__(clause, parents, task)
        if train_df is None:
            raise DatasetConfigError(
                'Attempt to create a memory dataset config, but train dataframe is not provided.'
            )
        self.train_df = train_df
        self.test_df = test_df
        self._train_cols = train_df.head(0)
        self._read_roles()

    def load_train(self) -> Dataset:
        retval = Dataset(metadata=self.metadata)
        retval.dataframe = self.train_df

        return retval

    def load_test(self) -> Optional[Dataset]:
        if self.test_df is None:
            return None

        retval = Dataset(metadata=self.metadata)
        retval.dataframe = self.test_df

        return self._prune_target(retval)


class MemoryDatasetFormat(DatasetFormat):
    '''Used to load a local tabular dataset.'''
    _builder = MemoryDatasetConfig
    _name = 'memory'
    _tags = {}


def register(catalog: DatasetFormatCatalog, *args, **kwargs):
    '''Register all the objects in this file.'''
    catalog.register(MemoryDatasetFormat(*args, **kwargs))



'''Metric object class for configurable and user-made metrics.'''
import abc

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional

from ...catalog.catalog_element_mixin import CatalogElementMixin
from ...wrangler.dataset import Dataset
from ...wrangler.dataset import DatasetKeys


class MetricError(Exception):
    '''Base error class for metrics'''


class MetricNameError(MetricError):
    '''Metric lacks a proper indexing name.'''


class MetricAttributeError(MetricError):
    '''Metric is missing a required attribute.'''


class MetricInvalidDatasetError(MetricError):
    '''Metric was passed a dataset it cannot use.'''


class Metric(CatalogElementMixin, metaclass=abc.ABCMeta):
    '''Base class wrapper for accessing metric objects.'''
    _name: Optional[str] = None
    _high: bool = True  # True means a higher score is better.
    _uses_proba: bool = False
    # True means metric uses class probabilites instead of class predictions.

    def __str__(self):
        return self.name

    @property
    def high(self) -> bool:
        '''Returns true if higher scores are better.'''
        return self._high

    @property
    def uses_proba(self) -> bool:
        '''True means metric uses class probabilites instead of class predictions.

        Should only be True for classification metrics.
        '''
        return self._uses_proba

    @abc.abstractmethod
    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        '''Calculates the metric results.'''


class SupervisedMetric(Metric):
    '''Subclass for supervised Metrics'''

    def validate_input(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> Dataset:
        '''Validates the Dataset and GroundTruth given. Useful for derived classes.'''
        if ground_truth is None:
            raise MetricInvalidDatasetError(
                'No Ground Truth Provided!')

        if DatasetKeys.PREDICTIONS.value not in pred:
            raise MetricInvalidDatasetError(
                f'Predictions has no "{DatasetKeys.PREDICTIONS.value}" key.')

        if DatasetKeys.GROUND_TRUTH.value not in ground_truth:
            raise MetricInvalidDatasetError(
                f'Ground truth has no "{DatasetKeys.GROUND_TRUTH.value}" key.')

        if ground_truth.metadata.target is None:
            raise MetricInvalidDatasetError(
                'Ground truth has no target column in metadata.')

        # We take a Optional[Dataset] and turn it into a Dataset. Type checkers are happy.
        return ground_truth

    @abc.abstractmethod
    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        '''
        Calculates the metric results.

        Most Supervised Metrics will require a ground truth. The implementation
        of the calculate should check if the ground truth exists.
        '''


class MetricStub(Metric):
    '''This is a stub'''
    _name = 'stub_metric'
    _high = True

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        return 0.0



'''Base class for sklearn models'''
# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=duplicate-code,too-many-arguments
from typing import Any, Dict, List, Optional
import importlib

import numpy as np
from numpy.random import RandomState
import pandas as pd

from ..impl.algorithm import Algorithm, AlgorithmCatalog
from ..impl.sklearn_like_algorithm_instance import SklearnLikeAlgorithmInstance
from ..impl.algorithm_instance import DatasetError
from ..impl.fittable_algorithm_instance import UntrainedError

from ...problem_def.task import DataType, TaskType
from ...wrangler.constants import ProblemDefKeys, Defaults
from ...wrangler.dataset import Dataset, DatasetKeys


def is_empty(frame) -> bool:
    '''Determine if an array-like object is empty.'''
    if frame is None:
        return True
    if isinstance(frame, np.ndarray) and frame.size == 0:
        # ndarray.size represents the number of elements in the array
        # ndarray.ndim represents the number of dimensions in the array
        #   arrays with more than 0 dimension can have 0 elements, and
        #   we consider those empty
        return True
    if isinstance(frame, pd.DataFrame) and frame.empty:
        return True
    return False


# TODO(Merritt): set random seed
class SklearnAlgorithm(Algorithm):
    '''Generic object for an sklearn algorithm.'''
    def __init__(self,
                 name: str,
                 algorithm: Optional[type] = None,
                 tasks: Optional[List[TaskType]] = None,
                 data_types: Optional[List[DataType]] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 **hyperparams: Any):
        self._name = name
        if algorithm is None:
            algorithm = self._load_module(name)
        self._algorithm = algorithm
        if tasks is None:
            tasks = []
        if data_types is None:
            data_types = []
        if tags is not None:
            self._tags = tags.copy()
        else:
            self._tags = {}
        self._tags.update({
            ProblemDefKeys.TASK_TYPE.value: [t.name.lower() for t in tasks],
            ProblemDefKeys.DATA_TYPE.value: [d.name.lower() for d in data_types],
        })
        super().__init__(**hyperparams)

    def _load_module(self, name: str):
        # Split name into module part (e.g. sklearn.linear_model)
        # and constructor part (e.g. LinearRegression)
        parts = name.split('.')
        constructor_part = parts[-1]
        module = importlib.import_module('.'.join(parts[:-1]))
        # Load the constructor.
        return getattr(module, constructor_part)

    def instantiate(self, **hyperparams) -> 'SklearnAlgorithmInstance':
        # TODO(Merritt): test that we're actually passing hyperparams through
        return SklearnAlgorithmInstance(
            parent=self,
            constructor=self._algorithm,
            **self._default_param_ranges(**self.hyperparams(**hyperparams)))


class SklearnAlgorithmInstance(SklearnLikeAlgorithmInstance):
    '''Generic object for an sklearn algorithm.'''

    def __init__(self, parent: Algorithm, constructor: type, *args, **kwargs):
        self._constructor = constructor
        assert parent._tags is not None, (
            'BUG: Algorithm __init__ should have set _tags'
        )
        if parent._tags.get('supports_random_seed', ['false'])[0] == 'true':
            if 'random_seed' in kwargs:
                kwargs['random_state'] = RandomState(kwargs.pop('random_seed'))
            elif kwargs.get('random_state', None) is None:
                kwargs['random_state'] = RandomState(Defaults.SEED)
        super().__init__(parent, *args, **kwargs)

    def fit(self, dataset: Dataset):
        assert DatasetKeys.TARGET.value in dataset, (
            "BUG: dataset.metadata.target is unexpectedly None")
        target_data = dataset[DatasetKeys.TARGET.value]
        if isinstance(target_data, pd.DataFrame):
            assert dataset.metadata.target is not None
            target_for_sklearn = list(
                target_data[dataset.metadata.target.name])
        else:
            target_for_sklearn = list(target_data)

        covariates = dataset[DatasetKeys.COVARIATES.value]

        new_dataset = dataset.output()
        new_dataset[DatasetKeys.TARGET.value] = target_for_sklearn
        new_dataset[DatasetKeys.COVARIATES.value] = covariates
        super().fit(dataset=new_dataset)

    def predict(self, dataset: Dataset) -> Dataset:
        if not self._trained:
            raise UntrainedError(f'attempt to predict before fit for {self._algorithm.name}')

        # this is not typed because it is supposed to hold the return value from sklearn
        # we expect it to be np.ndarray, but there is no guarantee in sklearn
        predictions_result = np.empty(0)
        probabilities_result = np.empty((0, 0))

        try:
            if not is_empty(dataset[DatasetKeys.COVARIATES.value]):
                predictions_result = self._impl.predict(dataset[DatasetKeys.COVARIATES.value])

                # assume the algorithm has a predict_proba function unless otherwise marked
                has_predict_proba = 'false' not in self._algorithm.tags.get(
                    'has_predict_proba', ['true'])
                if dataset.metadata.task == TaskType.BINARY_CLASSIFICATION and has_predict_proba:
                    probabilities_result = self._impl.predict_proba(
                        dataset[DatasetKeys.COVARIATES.value])

        except KeyError as err:
            raise DatasetError(f'predict dataset malformed {dataset!r}') from err

        # sklearn does not use typed Python, and thus we do not trust
        #    its return value to be consistently typed
        if is_empty(predictions_result):
            predictions_result = np.empty(0)

        retval = dataset.output()
        assert dataset.metadata.target is not None, (
            'BUG: predict() called with no target info in metadata'
        )
        retval.predictions = pd.DataFrame({
            dataset.metadata.target.name: predictions_result
        })

        if not is_empty(probabilities_result):
            # we arbitrarily choose only one column of probabilities to be included in retval
            probabilities_result = np.delete(probabilities_result, 0, axis=1).flatten()
            retval.probabilities = pd.DataFrame({
                dataset.metadata.target.name: probabilities_result
            })

        return retval


def register(catalog: AlgorithmCatalog):  # pylint: disable=unused-argument
    '''Nothing to register.

    All subclasses of SklearnAlgorithm are registered in sklearn_algorithms.py
    '''



'''Splitter that does random sampling.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..problem_def.cross_validation_config import CrossValidationConfig
from ..wrangler.dataset import Dataset
from ..wrangler.constants import Defaults

from .impl.splitter import Splitter, SplitterCatalog, SplitDataset, Fold


class SingleFoldSplitter(Splitter):
    '''Splitter that randomly splits data into one fold.

    It produces a single internal-train set and internal-validation set,
    according to the fraction "train_frac".
    '''
    _name = 'single_fold'
    _hyperparams = {
        'train_frac': Defaults.SPLIT_FRACTION
    }

    def split(self, dataset: Dataset, **overrides) -> SplitDataset:
        '''Split dataset into train and validation sets.

        We use pandas.DataFrame.sample to extract dataset_config.split_fraction
        as the 'train' dataset, and the rest as the 'validate' dataset.

        If there is a target role in the dataset, that column is removed from
        the 'validate' dataset and put in the 'ground_truth' dataset.
        '''
        train = dataset.output()
        validate = dataset.output()
        ground_truth = dataset.output()

        hyperparams = self.hyperparams(**overrides)
        train_frac = hyperparams['train_frac']

        data_df = dataset.get_dataframe()
        train_df = data_df.sample(frac=train_frac, random_state=self._cv_config.seed)
        validate_df = data_df.drop(train_df.index)

        if dataset.metadata.target is not None:
            assert dataset.metadata.target.name is not None
            ground_truth_df = validate_df[[str(dataset.metadata.target.name)]]
            ground_truth.ground_truth = ground_truth_df

            validate_df_no_target = validate_df.drop(dataset.metadata.target.name, axis=1)
        else:
            validate_df_no_target = validate_df

        train.dataframe = train_df
        validate.dataframe = validate_df_no_target

        return SplitDataset([Fold(train=train, validate=validate, ground_truth=ground_truth)])


def register(catalog: SplitterCatalog, *args, cv_config: CrossValidationConfig, **kwargs):
    '''Register all the objects in this file.'''
    catalog.register(SingleFoldSplitter(*args, cv_config=cv_config, **kwargs))
